/**
 * @file commsel.c
 * Communication protocol selection
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/*
 * Copyright  2001,2002 Marko Mkel
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>

#include "comm.h"
#include "commsel.h"
#ifdef COMM_SERIAL
# include "serial.h"
#endif /* COMM_SERIAL */
#ifdef COMM_PAR
# include "par.h"
#endif /* COMM_PAR */

/** Communication protocol bindings */
struct commproto
{
  /** name of the protocol */
  const char* name;
  /** array of bitmasks defining the pin assignments */
  const unsigned char* conn;
  /** connection establishment primitive
   * @param dev		name of the communication device
   * @param conn	array of bitmasks defining the pin assignments
   * @return		zero on success, nonzero on failure
   */
  int (*comm_init) (const char* dev,
		    const unsigned char* conn);
  /** connection termination primitive */
  void (*comm_fini) (void);
  /** the communication primitives */
  struct comm comm;
};

/** Supported communication protocols, starting from the default one */
static const struct commproto protocols[] =
{
#ifdef COMM_SERIAL
  {
    "c2n232", 0, serial_init, serial_fini,
    { 1, 1, 1, 1, serial_dtr, serial_rts, serial_txd, serial_cts }
  },
  {
    "dasa", 0, serial_init, serial_fini,
    { 0, 0, 0, 0, serial_rts, serial_dtr, serial_txd, serial_cts }
  },
  {
    "ponyprog", 0, serial_init, serial_fini,
    { 1, 0, 0, 0, serial_txd, serial_rts, serial_dtr, serial_cts }
  },
#endif /* COMM_SERIAL */
#ifdef COMM_PAR
  {
    "dapa",
    "\000\000\001" /* data bit masks for RESET, SCK, MOSI: nc, nc, D0 */
    "\004\001\000" /* ctrl bit masks for RESET, SCK, MOSI: INIT, STROBE, nc */
    "\200" /* status bit mask for MISO: BUSY */
    "\377\0" /* start-up state of the data and ctrl registers */
    "\377\4" /* clean-up state of the data and ctrl registers */,
    par_init, par_fini,
    { 0, 1, 0, 1, par_reset, par_sck, par_mosi, par_miso }
  },
  {
    "dt006",
    "\004\010\001" /* data bit masks for RESET, SCK, MOSI: D2, D3, D0 */
    "\000\000\000" /* ctrl bit masks for RESET, SCK, MOSI: none */
    "\200" /* status bit mask for MISO: BUSY */
    "\363\0" /* start-up state of the data and ctrl registers */
    "\363\0" /* clean-up state of the data and ctrl registers */,
    par_init, par_fini,
    { 0, 0, 0, 1, par_reset, par_sck, par_mosi, par_miso }
  },
  {
    "stk200" /* or stk300 */,
    "\300\020\040" /* data bit masks for RESET, SCK, MOSI: D7+D6, D4, D5 */
    "\000\000\000" /* ctrl bit masks for RESET, SCK, MOSI: none */
    "\100" /* status bit mask for MISO: ACK */
    "\343\0" /* start-up state of the data and ctrl registers */
    "\377\0" /* clean-up state of the data and ctrl registers */,
    par_init, par_fini,
    { 0, 0, 0, 0, par_reset, par_sck, par_mosi, par_miso }
  },
#endif /* COMM_PAR */
  /* end of list */
  { 0, 0, 0, 0, { 0, 0, 0, 0, 0, 0, 0, 0 } }
};

/** The connection cleanup function */
static void (*comm_fini) (void) = 0;

/** Initialize the communications of a protocol
 * @param protocol	the protocol
 * @param dev		name of the communication interface device
 * @return		pointer to the communication primitives, or NULL
 */
static const struct comm*
comm_init (const struct commproto* protocol,
	   const char* dev)
{
  int status = (*protocol->comm_init) (dev, protocol->conn);
  if (status) {
    fprintf (stderr, "init for '%s' returned %d\n", protocol->name, status);
    return 0;
  }
  comm_fini = protocol->comm_fini;
  return &protocol->comm;
}

/** Select a communication protocol by name, and initialize the communications
 * @param name		name of the protocol (NULL=default)
 * @param dev		name of the communication interface device
 * @param comm		(output) pointer to the communication primitives
 * @return		pointer to host information, or NULL on error
 */
const struct comm*
establish (const char* name,
	   const char* dev)
{
  const struct commproto* proto;
  if (!name)
    return comm_init (protocols, dev);
  for (proto = protocols; proto->name; proto++)
    if (!strcmp (proto->name, name))
      return comm_init (proto, dev);
  fprintf (stderr, "establish: no driver for '%s' found\n"
	   "Try one of the following:\n", name);
  for (proto = protocols; proto->name; proto++)
    fputs (proto->name, stderr), fputc ('\n', stderr);
  return 0;
}

/** Terminate the connection, if it was established */
void
terminate (void)
{
  if (!comm_fini)
    return;
  (*comm_fini) ();
  comm_fini = 0;
}
