/**
 * @file par.c
 * Parallel transfer routines
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/*
 * Copyright  2001,2002 Marko Mkel.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef COMM_PAR
# if defined USE_PARIO && !defined _BSD_SOURCE
#  define _BSD_SOURCE /* for seteuid(2) */
# endif /* _BSD_SOURCE */
# if defined __linux__ || defined __FreeBSD__
#  include <unistd.h>
#  include <sys/ioctl.h>
#  include <fcntl.h>
# else
#  undef USE_PARPORT
# endif /* __linux__ || __FreeBSD__ */
# if defined __linux__
#  ifdef USE_PARPORT
#   include <linux/ppdev.h>
#  endif /* USE_PARPORT */
#  ifdef USE_PARIO
#   include <sys/io.h>
#   define outbyte(value, port) outb (value, port)
#  endif /* USE_PARIO */
# endif /* __linux__ */
# if defined __FreeBSD__
#  ifdef USE_PARPORT
#   include <machine/sysarch.h>
#   include <dev/ppbus/ppi.h>
#  endif /* USE_PARPORT */
#  ifdef USE_PARIO
#   include <machine/cpufunc.h>
#   define outbyte(value, port) outb (port, value)
#  endif /* USE_PARIO */
# endif /* __FreeBSD__ */

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <errno.h>

/** data bit masks */
static unsigned char data_reset, data_sck, data_mosi;
/** control bit masks */
static unsigned char ctrl_reset, ctrl_sck, ctrl_mosi;
/** status bit mask for MISO */
static unsigned char stat_miso;
/** how to set the bits when the device is deactivated */
static unsigned char data_fini, ctrl_fini;

# ifdef USE_PARPORT
/** the parallel port file descriptor */
static int parfd = 0;
# endif /* USE_PARPORT */
# ifdef USE_PARIO
/** recognized base addresses for ports */
static const unsigned portbase[] = { 0x3bc, 0x378, 0x2bc, 0x278 };
/** the base address for the printer port, and the address of the data port */
static unsigned baseaddr = 0;
# if (defined __WIN32 || defined __WIN32__) && defined __GNUC__
/* copied from Linux <sys/io.h> */
static __inline unsigned char
inb (unsigned short int port)
{
  unsigned char _v;
  __asm__ __volatile__ ("inb %w1,%0":"=a" (_v):"Nd" (port));
  return _v;
}
static __inline void
outbyte (unsigned char value, unsigned short int port)
{
  __asm__ __volatile__ ("outb %b0,%w1": :"a" (value), "Nd" (port));
}
# endif /* __WIN32 || __WIN32__ */
# endif /* USE_PARIO */

/** state of the data and ctrl registers */
static unsigned char data, ctrl;

/** write data bits to the parallel port
 * @return	zero on success, nonzero on failure
 */
static int
write_data (void)
{
# ifdef USE_PARIO
  if (baseaddr) {
    outbyte (data, baseaddr);
    return 0;
  }
# endif /* USE_PARIO */
# ifdef USE_PARPORT
#  ifdef __linux__
  if (ioctl (parfd, PPWDATA, &data)) {
    perror ("ioctl (PPWDATA)");
    return -1;
  }
#  endif /* __linux__ */
#  ifdef __FreeBSD__
  if (ioctl (parfd, PPISDATA, &data)) {
    perror ("ioctl (PPISDATA)");
    return -1;
  }
#  endif /* __FreeBSD__ */
# endif /* USE_PARPORT */
  return 0;
}

/** write control bits to the parallel port
 * @return	zero on success, nonzero on failure
 */
static int
write_ctrl (void)
{
# ifdef USE_PARIO
  if (baseaddr) {
    outbyte (ctrl, baseaddr + 2);
    return 0;
  }
# endif /* USE_PARIO */
# ifdef USE_PARPORT
#  ifdef __linux__
  if (ioctl (parfd, PPWCONTROL, &ctrl)) {
    perror ("ioctl (PPWCONTROL)");
    return -1;
  }
#  endif /* __linux__ */
#  ifdef __FreeBSD__
  if (ioctl (parfd, PPISCTRL, &ctrl)) {
    perror ("ioctl (PPISCTRL)");
    return -1;
  }
#  endif /* __FreeBSD__ */
# endif /* USE_PARPORT */
  return 0;
}

/** Open the communication channel to direct parallel connection
 * @param dev	name of the communication device
 * @param conn	array of bitmasks defining the pin assignments
 * @return	zero on success, nonzero on failure
 */
int
par_init (const char* dev,
	  const unsigned char* conn)
{
# ifdef USE_PARIO
  char* s;
  unsigned long p = strtoul (dev, &s, 0);
  if (!*dev || *s) {
#  ifdef USE_PARPORT
    baseaddr = 0;
#  else /* USE_PARPORT */
    fprintf (stderr, "par_init: unrecognized port %s\n", dev);
    return 1;
#  endif /* USE_PARPORT */
  }
  else if (p < (sizeof portbase) / sizeof *portbase)
    baseaddr = portbase[p];
  else {
    unsigned i = (sizeof portbase) / sizeof *portbase;
    while (i)
      if (p == portbase[--i])
	goto found;
    fprintf (stderr, "par_init: port %s out of range\n", dev);
    return 1;
  found:
    baseaddr = portbase[i];
  }

  if (baseaddr) {
#  ifdef __linux__
    if (ioperm (baseaddr, 3, 1)) {
      perror ("par_init: ioperm");
      return 1;
    }
    /* give up the super user privileges */
    seteuid (getuid ());
#  endif /* __linux__ */
#  ifdef __FreeBSD__
    if (i386_set_ioperm (baseaddr, 3, 1)) {
      perror ("par_init: i386_set_ioperm");
      return 1;
    }
    /* give up the super user privileges */
    seteuid (getuid ());
#  endif /* __FreeBSD__ */
  }
#  ifdef USE_PARPORT
  else {
#  endif /* USE_PARPORT */
# endif /* USE_PARIO */
# ifdef USE_PARPORT
    if (0 > (parfd = open (dev, O_RDWR, 0))) {
      fputs ("par_init ", stderr);
      fputs(dev, stderr);
      perror (": open");
      return 1;
    }
#  ifdef __linux__
    if (ioctl (parfd, PPCLAIM, 0)) {
      fputs ("par_init ", stderr);
      fputs(dev, stderr);
      perror (": ioctl (PPCLAIM)");
      close (parfd);
      parfd = 0;
      return 1;
    }
#  endif /* __linux */
#  ifdef USE_PARIO
  }
#  endif /* USE_PARIO */
# endif /* USE_PARPORT */
  data_reset = *conn++, data_sck = *conn++, data_mosi = *conn++;
  ctrl_reset = *conn++, ctrl_sck = *conn++, ctrl_mosi = *conn++;
  stat_miso = *conn++;
  data = *conn++, ctrl = *conn++, data_fini = *conn++, ctrl_fini = *conn++;

  if (write_ctrl () || write_data ())
    return 2;

  return 0;
}

/** Close the communication channel to direct parallel connection */
void
par_fini (void)
{
  data = data_fini, ctrl = ctrl_fini;
# ifdef USE_PARIO
  if (baseaddr) {
    write_data (), write_ctrl ();
    baseaddr = 0;
  }
# endif /* USE_PARIO*/
# ifdef USE_PARPORT
  if (parfd > 0) {
    write_data (), write_ctrl ();
#  ifdef __linux__
    ioctl (parfd, PPRELEASE, 0);
#  endif /* __linux__ */
    close (parfd);
    parfd = 0;
  }
# endif /* USE_PARPORT */
}

/** Define a routine for controlling a signal
 * @param sig	the signal to be controlled
 */
#define control(sig)						\
int								\
par_##sig (int bit)						\
{								\
  int st = -1;							\
  if (data_##sig) {						\
    if (bit) data |= data_##sig; else data &= ~data_##sig;	\
    if ((st = write_data ()))					\
      return st;						\
  }								\
  if (ctrl_##sig) {						\
    if (bit) ctrl |= ctrl_##sig; else ctrl &= ~ctrl_##sig;	\
    if ((st = write_ctrl ()))					\
      return st;						\
  }								\
  return st;							\
}

control (reset)
control (sck)
control (mosi)
#undef control

/** Read MISO
 * @return	0 (negated), 1 (asserted), or an error code
 */
int
par_miso (void)
{
  unsigned char stat;
# ifdef USE_PARIO
  if (baseaddr)
    stat = inb (baseaddr + 1);
#  ifndef USE_PARPORT
  else
    return -1;
#  else /* !USE_PARPORT */
  else {
#  endif /* !USE_PARPORT */
# endif /* USE_PARIO */
# ifdef USE_PARPORT
#  ifdef __linux__
    if (ioctl (parfd, PPRSTATUS, &stat)) {
      perror ("ioctl (PPRSTATUS)");
      return -1;
    }
#  endif /* __linux__ */
#  ifdef __FreeBSD__
    if (ioctl (parfd, PPIGSTATUS, &stat)) {
      perror ("ioctl (PPIGSTATUS)");
      return -1;
    }
#  endif /* __FreeBSD__ */
#  ifdef USE_PARIO
  }
#  endif /* USE_PARIO */
# endif /* USE_PARPORT */
  return stat & stat_miso ? 1 : 0;
}

#endif /* COMM_PAR */
