/**
 * @file ihex.c
 * Intel hex input and output
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/* Copyright  2002 Marko Mkel.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <ctype.h>
#include "comm.h"
#include "ihex.h"

/** Construct hex dump from source
 * @param file	the output file
 * @param comm	the communication primitives
 * @param avr	the device descriptor
 * @param len	number of bytes to dump (start from zero)
 * @param src	the data source
 * @return	zero on success; nonzero on error
 */
int
ihex_output (FILE* file,
	     const struct comm* comm,
	     const struct avr* avr,
	     unsigned len,
	     int (*src) (const struct comm*, const struct avr*,
			 unsigned, unsigned char*))
{
  unsigned rec;
  for (rec = 0; rec <= len; rec += 16) {
    unsigned first = rec, last = rec + 16 > len ? len : rec + 16, cksum =
      (last - first) + (first & 0xff) + (first >> 8) + (first == last);
    fprintf (file, ":%02X%04X%02X", last - first, first, first == last);

    for (; first < last; first++) {
      unsigned char ch;
      int st = (*src) (comm, avr, first, &ch);
      if (st)
	return st;
      fprintf (file, "%02X", ch);
      cksum += ch;
    }

    fprintf (file, "%02X\r\n", (-cksum) & 0xff);
  }

  return 0;
}

/** Read a hexadecimal digit
 * @param file	the input file
 * @return	the digit, or -1 on error
 */
static int
gethexdigit (FILE* file)
{
  int c = getc (file);
  if (c == EOF) {
    fputs ("ihex_input: gethexdigit: unexpected EOF\n", stderr);
    return -1;
  }
  if (c >= '0' && c <= '9')
    return c - '0';
  if (c >= 'A' && c <= 'F')
    return 10 + (c - 'A');
  if (c >= 'a' && c <= 'f')
    return 10 + (c - 'a');
  fputs ("ihex_input: unrecognized hexadecimal digit\n", stderr);
  return -1;
}

/** Read a two-digit hexadecimal number
 * @param file	the input file
 * @return	the number, or -1 on error
 */
static int
gethex (FILE* file)
{
  int c1, c2;
  if (-1 == (c1 = gethexdigit (file)) ||
      -1 == (c2 = gethexdigit (file)))
    return -1;
  return c1 << 4 | c2;
}

/** Write to dest from hex dump
 * @param file	the input file
 * @param comm	the communication primitives
 * @param avr	the device descriptor
 * @param dest	the data destination
 * @param done	function to invoke at the end of the dump (0=none)
 * @return	zero on success; nonzero on error
 */
int
ihex_input (FILE* file,
	    const struct comm* comm,
	    const struct avr* avr,
	    int (*dest) (const struct comm*, const struct avr*,
			 unsigned, unsigned char),
	    int (*done) (const struct comm*, const struct avr*))
{
  for (;;) {
    unsigned length, first, last, cksum;
    /* awaiting colon */
    int c = getc (file);
    if (c == EOF) {
      fputs ("ihex_input: unexpected EOF\n", stderr);
      return -1;
    }
    if (isspace (c))
      continue;
    if (c != ':') {
      fputs ("ihex_input: unexpected character at start of line\n", stderr);
      return -1;
    }
    /* length of this line */
    if (-1 == (c = gethex (file)))
      return -1;
    cksum = c;
    length = c;
    if (-1 == (c = gethex (file)))
      return -1;
    cksum += c;
    first = c;
    if (-1 == (c = gethex (file)))
      return -1;
    cksum += c;
    first = (first << 8) | c;
    last = first + length;
    if (-1 == (c = gethex (file)))
      return -1;
    cksum += c;
    if (c != !length) {
      fputs ("ihex_input: unexpected EOF flag value\n", stderr);
      return -1;
    }
    for (; first < last; first++) {
      if (-1 == (c = gethex (file)))
	return -1;
      cksum += c;
      if ((*dest) (comm, avr, first, (unsigned char) c))
	return -2;
    }
    if (-1 == (c = gethex (file)))
      return -1;
    cksum += c;
    if (cksum & 0xff) {
      fputs ("ihex_input: checksum mismatch\n", stderr);
      return -1;
    }
    if (!length)
      return done ? (*done) (comm, avr) : 0;
  }
}
