/**
 * @file spi.c
 * Serial mode programming interface for AVR microcontrollers
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/* Copyright  2002 Marko Mkel.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _POSIX_C_SOURCE 199309L
#define _POSIX_VERSION 199309L
#include "comm.h"
#include "spi.h"
#include <stdio.h>

#if defined __unix || defined __unix__ || defined __APPLE__
# ifndef unix
#  define unix
# endif
# include <sys/time.h>
# include <time.h>
#endif /* defined __unix || defined __unix__ || defined __APPLE__ */

#if defined __WIN32 || defined __WIN32__
# include <windows.h>
# ifndef WIN32
#  define WIN32
# endif /* WIN32 */
#endif /* defined __WIN32 || defined __WIN32__ */

/** Millisecond delay
 * @param msec	delay time in milliseconds
 */
static void
msecdelay (unsigned msec)
{
#ifdef WIN32
  Sleep (msec);
#endif /* WIN32 */
#ifdef unix
  struct timespec time;
  time.tv_sec = 0;
  time.tv_nsec = 1000000L * msec;
  nanosleep (&time, 0);
#endif /* unix */
}

/** Delay for minimum SCK high/low time (2 s) */
static void
sckdelay (void)
{
#ifdef unix
  struct timeval t1, t2;
  gettimeofday (&t1, 0);
  if ((t1.tv_usec += 2) >= 1000000L)
    t1.tv_usec -= 1000000L, t1.tv_sec++;
  do
    gettimeofday (&t2, 0);
  while (t1.tv_sec == t2.tv_sec
	 ? t1.tv_usec > t2.tv_usec
	 : t1.tv_sec > t2.tv_sec);
#endif
}

/** Send and receive a byte
 * @param comm		the communication primitives
 * @param c		the byte to be sent
 * @param output	(out) the received byte
 * @return		zero on success, nonzero on failure
 */
static int
sr (const struct comm* comm,
    unsigned char c,
    unsigned char* output)
{
  unsigned char mask = 0x80, ch = 0;
  for (; mask; mask >>= 1) {
    if ((*comm->comm_mosi) (c & mask ? !comm->fInvMOSI : comm->fInvMOSI))
      return -1;
    sckdelay ();
    switch ((*comm->comm_miso) ()) {
    case 1:
      if (!comm->fInvMISO) ch |= mask;
      break;
    case 0:
      if (comm->fInvMISO) ch |= mask;
      break;
    default:
      return -3;
    }
    if ((*comm->comm_sck) (!comm->fInvSCK))
      return -2;
    sckdelay ();
    if ((*comm->comm_sck) (comm->fInvSCK))
      return -4;
  }

  *output = ch;
  return 0;
}

/** Send and receive 4 bytes of data
 * @param comm		the communication primitives
 * @param input		the data to be sent
 * @param output	the received data
 * @param context	if non-null, complain if output does not match input
 * @return		zero on success, nonzero on failure
 */
int
sendrecv (const struct comm* comm,
	  const unsigned char* input,
	  unsigned char* output,
	  const char* context)
{
  unsigned len = 4;
  while (len--) {
    int st;
#if DEBUG
    fprintf (stderr, "%02x", *input);
#endif
    if ((st = sr (comm, *input++, output++)))
      return st;
#if DEBUG
    fprintf (stderr, "(%02x)", output[-1]);
#endif
  }
#if DEBUG
  fputc ('\n', stderr);
#endif
  if (context && (output[-3] != input[-4] || output[-2] != input[-3])) {
    fprintf (stderr, "%s: unexpected response %02x %02x %02x %02x"
	     " to %02x %02x %02x %02x\n", context,
	     input[-4], input[-3], input[-2], input[-1],
	     output[-4], output[-3], output[-2], output[-1]);
    return -1;
  }
  return 0;
}

/** Enable the programming mode
 * @param comm		the communication primitives
 * @return		zero on success, nonzero on failure
 */
int
enable (const struct comm* comm)
{
  static const unsigned char prg[4] = { 0xac, 0x53, 0, 0 };
  unsigned char resp[4];
  unsigned retries = 32;
  if ((*comm->comm_reset) (comm->fInvRESET))
    return -1;
  if ((*comm->comm_sck) (comm->fInvSCK))
    return -2;
  while (retries--) {
    int st;
    if ((st = sendrecv (comm, prg, resp, 0)))
      return st;
    if (resp[2] == 0x53)
      return 0;
    /* pulse the SCK line */
    if ((st = (*comm->comm_sck) (!comm->fInvSCK)))
      return st;
    sckdelay ();
    if ((st = (*comm->comm_sck) (comm->fInvSCK)))
      return st;
    sckdelay ();
  }

  return -1;
}

/** Get device information
 * @param comm		the communication primitives
 * @param ch		(in) information to look up, (out) the read data
 * @return		zero on success, nonzero on failure
 */
static int
deviceinfo (const struct comm* comm,
	    unsigned char* ch)
{
  unsigned char info[4], resp[4];
  int st;
  info[0] = 0x30, info[2] = *ch, info[1] = info[3] = 0;
  if ((st = sendrecv (comm, info, resp, "deviceinfo")))
    return st;
  *ch = resp[3];
  return 0;
}

/** Table of supported devices.
 * Adapted from uisp-20020303, files Avr.h and Avr.C
 */
static const struct avr devices[] =
  {
    /* signature  name          flash page EEPROM fd ed bp pp */
    { 0x90, 0x01, "AT90S1200",   1024,  0,    64, 4, 4, 0, 0 },
    { 0x90, 0x05, "ATtiny12",    1024,  0,    64, 2, 4, 1, 0 },
    { 0x90, 0x06, "ATtiny15",    1024,  0,    64, 2, 4, 1, 0 },
    { 0x91, 0x01, "AT90S2313",   2048,  0,   128, 4, 4, 1, 0 },
    { 0x91, 0x02, "AT90S2343",   2048,  0,   128, 4, 4, 1, 0 },
    { 0x91, 0x03, "AT90S2323",   2048,  0,   128, 4, 4, 1, 0 },
    { 0x91, 0x05, "AT90S2333",   2048,  0,   128, 4, 4, 1, 0 },
    { 0x91, 0x06, "ATtiny22",    2048,  0,   128, 4, 4, 1, 0 },
    { 0x92, 0x01, "AT90S4414",   4096,  0,   256, 4, 4, 1, 0 },
    { 0x92, 0x02, "AT90S4434",   4096,  0,   256, 4, 4, 1, 0 },
    { 0x92, 0x03, "AT90S4433",   4096,  0,   256, 4, 4, 1, 0 },
    { 0x93, 0x01, "AT90S8515",   8192,  0,   512, 4, 4, 1, 0 },
    { 0x93, 0x03, "AT90S8535",   8192,  0,   512, 4, 4, 1, 0 },
    { 0x93, 0x07, "ATmega8",     8192,  64,  512, 5, 9, 1, 1 },
    { 0x94, 0x01, "ATmega161",  16384, 128,  512, 11, 4, 1, 1 },
    { 0x94, 0x02, "ATmega163",  16384, 128,  512, 15, 4, 1, 1 },
    { 0x94, 0x03, "ATmega16",   16384, 128,  512, 5, 9, 1, 1 },
    { 0x95, 0x01, "ATmega323",  32768, 128, 1024, 15, 4, 1, 1 },
    { 0x96, 0x02, "ATmega64",   65536, 256, 2048, 5, 9, 1, 1 },
    { 0x97, 0x01, "ATmega103", 131072, 256, 4096, 22, 4, 1, 0 },
    { 0x97, 0x02, "ATmega128", 131072, 256, 4096, 5, 9, 1, 1 },
    { 0x01, 0x01, "ATmega103o",131072, 256, 4096, 22, 4, 1, 0 }
  };

/** Detect the AVR
 * @param comm		the communication primitives
 * @return		pointer to the device descriptor, or 0 on failure
 */
const struct avr*
detect (const struct comm* comm)
{
  unsigned char info[3];
  int st;
  info[0] = 0, info[1] = 1, info[2] = 2;
  if ((st = deviceinfo (comm, &info[0])) ||
      (st = deviceinfo (comm, &info[1])) ||
      (st = deviceinfo (comm, &info[2])))
    return 0;
  if (info[0] == 0x1e) {
    unsigned i = (sizeof devices) / sizeof *devices;
    while (i--)
      if (info[1] == devices[i].family && info[2] == devices[i].number)
	return devices + i;
  }
  else if (info[0] == 0 && info[1] == 1 && info[2] == 2) {
    fputs ("detect: device is locked\n", stderr);
    return 0;
  }

  fprintf (stderr, "detect: unknown signature %02x %02x %02x\n",
	   info[0], info[1], info[2]);
  return 0;
}

/** Erase the device
 * @param comm		the communication primitives
 * @return		zero on success, nonzero on failure
 */
int
erase (const struct comm* comm)
{
  static const unsigned char nuke[4] = { 0xac, 0x80, 0x00, 0x00 };
  unsigned char resp[4];
  int st = sendrecv (comm, nuke, resp, 0);
  msecdelay (100); /* twice the longest flash_delay */
  return st;
}

/** Read a byte from flash
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		(output) the read byte
 * @return		zero on success, nonzero on failure
 */
int
read_flash (const struct comm* comm, const struct avr* avr,
	    unsigned addr, unsigned char* ch)
{
  unsigned char cmd[4], resp[4];
  int st;
  if (addr >= avr->flash_size) {
    fprintf (stderr, "read_flash: address %04x out of range 0..%04x\n",
	     addr, avr->flash_size);
    return -1;
  }
  cmd[0] = addr & 1 ? 0x28 : 0x20;
  cmd[1] = addr >> 9;
  cmd[2] = addr >> 1;
  cmd[3] = 0;
  if ((st = sendrecv (comm, cmd, resp, "read_flash")))
    return st;
  *ch = resp[3];
  return 0;
}

/** Read a byte from EEPROM
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		(output) the read byte
 * @return		zero on success, nonzero on failure
 */
int
read_eeprom (const struct comm* comm, const struct avr* avr,
	     unsigned addr, unsigned char* ch)
{
  unsigned char cmd[4], resp[4];
  int st;
  if (addr >= avr->eeprom_size) {
    fprintf (stderr, "read_eeprom: address %04x out of range 0..%04x\n",
	     addr, avr->eeprom_size);
    return -1;
  }
  cmd[0] = 0xa0;
  cmd[1] = addr >> 8;
  cmd[2] = addr;
  cmd[3] = 0;
  if ((st = sendrecv (comm, cmd, resp, "read_eeprom")))
    return st;
  *ch = resp[3];
  return 0;
}

/** Verify a byte with flash
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		the byte to verify
 * @return		zero on success, nonzero on failure
 */
int
verify_flash (const struct comm* comm, const struct avr* avr,
	      unsigned addr, unsigned char ch)
{
  unsigned char orig;
  int st = read_flash (comm, avr, addr, &orig);
  if (st) return st;
  if (orig != ch)
    fprintf (stderr, "verify_flash: %04x contains %02x instead of %02x\n",
	     addr, orig, ch);
  return 0;
}

/** Verify a byte with EEPROM
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		the byte to verify
 * @return		zero on success, nonzero on failure
 */
int
verify_eeprom (const struct comm* comm, const struct avr* avr,
	       unsigned addr, unsigned char ch)
{
  unsigned char orig;
  int st = read_eeprom (comm, avr, addr, &orig);
  if (st) return st;
  if (orig != ch)
    fprintf (stderr, "verify_eeprom: %04x contains %02x instead of %02x\n",
	     addr, orig, ch);
  return 0;
}

/** address of the flash page being written */
static unsigned page_addr;
/** address of the last programmed byte on the flash page */
static unsigned page_prog_addr;
/** last programmed byte on the flash page (0xff=none) */
static unsigned char page_prog_byte = 0xff;

/** Write a page to flash
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @return		zero on success, nonzero on failure
 */
static int
finish_page (const struct comm* comm, const struct avr* avr)
{
  int st;
  unsigned char cmd[4], resp[4];
#ifdef unix
  struct timeval t1, t2;
#endif
  cmd[0] = 0x4c, cmd[1] = page_addr >> 9, cmd[2] = page_addr >> 1, cmd[3] = 0;
  if ((st = sendrecv (comm, cmd, resp, "write_flash_page")))
    return st;
#ifdef unix
  gettimeofday (&t1, 0);
  if ((t1.tv_usec += 1000L * avr->flash_delay) >= 1000000L)
    t1.tv_usec -= 1000000L, t1.tv_sec++;
#endif /* unix */
  for (;;) {
#ifdef unix
    gettimeofday (&t2, 0);
    if (t1.tv_sec == t2.tv_sec
	? t1.tv_usec < t2.tv_usec
	: t1.tv_sec < t2.tv_sec)
      break;
#endif
    if (avr->fPagePoll && page_prog_byte != 0xff) {
      if ((st = read_flash (comm, avr, page_prog_addr, cmd)))
	return st;
      if (cmd[0] == page_prog_byte)
	break;
    }
    msecdelay (avr->flash_delay);
#ifdef WIN32
    break;
#endif /* WIN32 */
  }
  page_prog_byte = 0xff;
  return 0;
}

/** Write a byte to flash
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		the byte to write
 * @return		zero on success, nonzero on failure
 */
int
write_flash (const struct comm* comm, const struct avr* avr,
	     unsigned addr, unsigned char ch)
{
  unsigned char cmd[4], resp[4];
  if (avr->flash_page) {
    if ((page_addr ^ addr) & ~(avr->flash_page - 1)) {
      int st = finish_page (comm, avr);
      if (st) return st;
      page_addr = addr;
    }
    page_prog_addr = addr, page_prog_byte = ch;
    cmd[0] = addr & 1 ? 0x48 : 0x40;
    cmd[1] = addr >> 9;
    cmd[2] = addr >> 1;
    cmd[3] = ch;
    return sendrecv (comm, cmd, resp, "write_flash");
  }
  else {
    unsigned char orig;
    int st = read_flash (comm, avr, addr, &orig);
    if (st)
      return st;
    if (ch == orig)
      return 0;
    if (ch & ~orig) {
      fprintf (stderr, "write_flash: %04x contains %02x;"
	       " cannot write %02x without erasing\n",
	       addr, orig, ch);
      return -1;
    }
    cmd[0] = addr & 1 ? 0x48 : 0x40;
    cmd[1] = addr >> 9;
    cmd[2] = addr >> 1;
    cmd[3] = ch;
    if ((st = sendrecv (comm, cmd, resp, "write_flash")))
      return st;
    if (!avr->fBytePoll || !(ch & 0x7f) || (ch & 0x7f) == 0x7f)
      msecdelay (avr->flash_delay);
    else {
      unsigned char ch2;
#ifdef unix
      struct timeval t1, t2;
      gettimeofday (&t1, 0);
      if ((t1.tv_usec += 1000 * avr->flash_delay) >= 1000000L)
	t1.tv_usec -= 1000000L, t1.tv_sec++;
#endif
      for (;;) {
	if ((st = read_flash (comm, avr, addr, &ch2)))
	  return st;
	if (ch2 != orig && ch2 != 0x7f)
	  break;
#ifdef unix
	gettimeofday (&t2, 0);
	if (t1.tv_sec == t2.tv_sec
	    ? t1.tv_usec < t2.tv_usec
	    : t1.tv_sec < t2.tv_sec)
	  break;
#endif
      }
    }
    if ((st = read_flash (comm, avr, addr, &orig)))
      return st;
    if (ch == orig)
      return 0;
    fprintf (stderr, "write_flash: writing %02x to %04x failed: got %02x\n",
	     addr, ch, orig);
    return -1;
  }
}

/** Flush the last flash byte
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @return		zero on success, nonzero on failure
 */
int
flush_flash (const struct comm* comm,
	     const struct avr* avr)
{
  int st = avr->flash_page ? finish_page (comm, avr) : 0;
  page_addr = 0;
  return st;
}

/** Write a byte to EEPROM
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		the byte to write
 * @return		zero on success, nonzero on failure
 */
int
write_eeprom (const struct comm* comm, const struct avr* avr,
	      unsigned addr, unsigned char ch)
{
  int st;
  unsigned char cmd[4], resp[4];
  unsigned char orig;
  st = read_eeprom (comm, avr, addr, &orig);
  if (st)
    return st;
  if (ch == orig)
    return 0;
  if (ch & ~orig) {
    fprintf (stderr, "write_eeprom: %04x contains %02x;"
	     " cannot write %02x without erasing\n",
	     addr, orig, ch);
    return -1;
  }
  cmd[0] = addr & 1 ? 0x48 : 0x40;
  cmd[1] = addr >> 9;
  cmd[2] = addr >> 1;
  cmd[3] = ch;
  if ((st = sendrecv (comm, cmd, resp, "write_eeprom")))
    return st;
  if (!avr->fBytePoll || !(ch & 0x7f) || (ch & 0x7f) == 0x7f)
    msecdelay (avr->eeprom_delay);
  else {
    unsigned char ch2;
#ifdef unix
    struct timeval t1, t2;
    gettimeofday (&t1, 0);
    if ((t1.tv_usec += 1000 * avr->eeprom_delay) >= 1000000L)
      t1.tv_usec -= 1000000L, t1.tv_sec++;
#endif
    for (;;) {
      if ((st = read_eeprom (comm, avr, addr, &ch2)))
	return st;
      if (ch2 != orig && (ch2 & 0x7f) && (ch2 & 0x7f) != 0x7f)
	break;
#ifdef unix
      gettimeofday (&t2, 0);
      if (t1.tv_sec == t2.tv_sec
	  ? t1.tv_usec < t2.tv_usec
	  : t1.tv_sec < t2.tv_sec)
	break;
#endif
    }
  }
  if ((st = read_eeprom (comm, avr, addr, &orig)))
    return st;
  if (ch == orig)
    return 0;
  fprintf (stderr, "write_eeprom: writing %02x to %04x failed: got %02x\n",
	   addr, ch, orig);
  return -1;
}
