/**
 * @file spi.h
 * Serial mode programming interface for AVR microcontrollers
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/* Copyright  2002 Marko Mkel.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** Send and receive 4 bytes of data
 * @param comm		the communication primitives
 * @param input		the data to be sent
 * @param output	the received data
 * @param context	if non-null, complain if output does not match input
 * @return		zero on success, nonzero on failure
 */
int
sendrecv (const struct comm* comm,
	  const unsigned char* input,
	  unsigned char* output,
	  const char* context);

/** Enable the programming mode
 * @param comm		the communication primitives
 * @return		zero on success, nonzero on failure
 */
int
enable (const struct comm* comm);

/** Detect the AVR
 * @param comm		the communication primitives
 * @return		pointer to the device descriptor, or 0 on failure
 */
const struct avr*
detect (const struct comm* comm);

/** Erase the device
 * @param comm		the communication primitives
 * @return		zero on success, nonzero on failure
 */
int
erase (const struct comm* comm);

/** Read a byte from flash
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		(output) the read byte
 * @return		zero on success, nonzero on failure
 */
int
read_flash (const struct comm* comm, const struct avr* avr,
	    unsigned addr, unsigned char* ch);

/** Read a byte from EEPROM
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		(output) the read byte
 * @return		zero on success, nonzero on failure
 */
int
read_eeprom (const struct comm* comm, const struct avr* avr,
	     unsigned addr, unsigned char* ch);

/** Verify a byte with flash
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		the byte to verify
 * @return		zero on success, nonzero on failure
 */
int
verify_flash (const struct comm* comm, const struct avr* avr,
	      unsigned addr, unsigned char ch);

/** Verify a byte with EEPROM
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		the byte to verify
 * @return		zero on success, nonzero on failure
 */
int
verify_eeprom (const struct comm* comm, const struct avr* avr,
	       unsigned addr, unsigned char ch);

/** Write a byte to flash
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		the byte to write
 * @return		zero on success, nonzero on failure
 */
int
write_flash (const struct comm* comm, const struct avr* avr,
	     unsigned addr, unsigned char ch);

/** Flush the last flash byte
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @return		zero on success, nonzero on failure
 */
int
flush_flash (const struct comm* comm,
	     const struct avr* avr);

/** Write a byte to EEPROM
 * @param comm		the communication primitives
 * @param avr		the device descriptor
 * @param addr		the address
 * @param ch		the byte to write
 * @return		zero on success, nonzero on failure
 */
int
write_eeprom (const struct comm* comm, const struct avr* avr,
	      unsigned addr, unsigned char ch);
