/**
 * @file srec.c
 * Motorola S record input and output
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/* Copyright  2002 Marko Mkel.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <ctype.h>
#include "comm.h"
#include "srec.h"

/** Construct hex dump from source
 * @param file	the output file
 * @param comm	the communication primitives
 * @param avr	the device descriptor
 * @param len	number of bytes to dump (start from zero)
 * @param src	the data source
 * @return	zero on success; nonzero on error
 */
int
srec_output (FILE* file,
	     const struct comm* comm,
	     const struct avr* avr,
	     unsigned len,
	     int (*src) (const struct comm*, const struct avr*,
			 unsigned, unsigned char*))
{
  unsigned rec;
  for (rec = 0; rec < len; rec += 16) {
    unsigned first = rec, last = rec + 16 > len ? len : rec + 16, len, cksum =
      (first & 0xff) + (first >> 8);
    if (first < 0x10000) {
      fprintf (file, "S1%02X%04X", len = last - first + 3, first);
      cksum += len;
    }
    else if (first < 0x1000000) {
      fprintf (file, "S2%02X%06X", len = last - first + 4, first);
      cksum += len + (first >> 16);
    }
    else {
      fprintf (file, "S3%02X%08X", len = last - first + 5, first);
      cksum += len + (first >> 16) + (first >> 24);
    }

    for (; first < last; first++) {
      unsigned char ch;
      int st = (*src) (comm, avr, first, &ch);
      if (st)
	return st;
      fprintf (file, "%02X", ch);
      cksum += ch;
    }

    fprintf (file, "%02X\r\n", (~cksum) & 0xff);
  }

  fputs ("S9030000FC\r\n", file);
  return 0;
}

/** Read a hexadecimal digit
 * @param file	the input file
 * @return	the digit, or -1 on error
 */
static int
gethexdigit (FILE* file)
{
  int c = getc (file);
  if (c == EOF) {
    fputs ("srec_input: gethexdigit: unexpected EOF\n", stderr);
    return -1;
  }
  if (c >= '0' && c <= '9')
    return c - '0';
  if (c >= 'A' && c <= 'F')
    return 10 + (c - 'A');
  if (c >= 'a' && c <= 'f')
    return 10 + (c - 'a');
  fputs ("srec_input: unrecognized hexadecimal digit\n", stderr);
  return -1;
}

/** Read a two-digit hexadecimal number
 * @param file	the input file
 * @return	the number, or -1 on error
 */
static int
gethex (FILE* file)
{
  int c1, c2;
  if (-1 == (c1 = gethexdigit (file)) ||
      -1 == (c2 = gethexdigit (file)))
    return -1;
  return c1 << 4 | c2;
}

/** Write to dest from hex dump
 * @param file	the input file
 * @param comm	the communication primitives
 * @param avr	the device descriptor
 * @param dest	the data destination
 * @param done	function to invoke at the end of the dump (0=none)
 * @return	zero on success; nonzero on error
 */
int
srec_input (FILE* file,
	    const struct comm* comm,
	    const struct avr* avr,
	    int (*dest) (const struct comm*, const struct avr*,
			 unsigned, unsigned char),
	    int (*done) (const struct comm*, const struct avr*))
{
  for (;;) {
    unsigned length, first, last, cksum;
    /* awaiting 'S' */
    int c = getc (file), t;
    if (c == EOF) {
      fputs ("srec_input: unexpected EOF\n", stderr);
      return -1;
    }
    if (isspace (c))
      continue;
    if (c != 'S') {
      fputs ("srec_input: unexpected character at start of line\n", stderr);
      return -1;
    }
    /* type of the record */
    t = getc (file);
    /* length of the record */
    if (-1 == (c = gethex (file)))
      return -1;
    cksum = c + 1;
    length = c;
    if (length < 3) {
      fprintf (stderr, "srec_input: too short record (%u<3)\n", length);
      return -1;
    }
    length -= 3;
    if (-1 == (c = gethex (file)))
      return -1;
    cksum += c;
    first = c;
    if (-1 == (c = gethex (file)))
      return -1;
    cksum += c;
    first = (first << 8) | c;
    switch (t) {
    case '0':
      do
	if (-1 == (c = gethex (file)))
	  return -1;
      while (cksum += c, length--);
      goto sum;
    case '3':
      if (-1 == (c = gethex (file)))
	return -1;
      cksum += c;
      first = (first << 8) | c;
      length--;
      /* fall through */
    case '2':
      if (-1 == (c = gethex (file)))
	return -1;
      cksum += c;
      first = (first << 8) | c;
      length--;
      /* fall through */
    case '1':
      break;

    case '7':
      if (-1 == (c = gethex (file)))
	return -1;
      cksum += c;
      first = (first << 8) | c;
      length--;
      /* fall through */
    case '8':
      if (-1 == (c = gethex (file)))
	return -1;
      cksum += c;
      first = (first << 8) | c;
      length--;
      /* fall through */
    case '9':
      if (first)
	fprintf (stderr, "srec_input: ignoring start address 0x%X\n", first);
      if (-1 == (c = gethex (file)))
	return -1;
      cksum += c;
      if (length) {
	fputs ("srec_input: length mismatch\n", stderr);
	return -1;
      }
      if (cksum & 0xff) {
	fputs ("srec_input: checksum mismatch\n", stderr);
	return -1;
      }
      goto done;
    }

    for (last = first + length; first < last; first++) {
      if (-1 == (c = gethex (file)))
	return -1;
      cksum += c;
      if ((*dest) (comm, avr, first, (unsigned char) c))
	return -2;
    }
    if (-1 == (c = gethex (file)))
      return -1;
    cksum += c;
  sum:
    if (cksum & 0xff) {
      fputs ("srec_input: checksum mismatch\n", stderr);
      return -1;
    }
  }

 done:
  return done ? (*done) (comm, avr) : 0;
}
